{'config-search-paths': ['{:user-config-path}/idlib/config.yaml',],
 'auth-variables':
 {'cache-path': {'default': '{:user-cache-path}/idlib',
                 'environment-variables': 'IDLIB_CACHE_PATH CACHE_PATH'},
  'log-path': {'default': '{:user-log-path}/idlib',
               'environment-variables': 'IDLIB_LOG_PATH LOG_PATH'},
  'protocols-io-api-creds-file': None,
  'protocols-io-api-store-file': None,
 }}
