from .ark import Ark
from .doi import Doi
from .handle import Handle
from .orcid import Orcid
from .pmid import Pmid
from .ror import Ror
from .rrid import Rrid
from .urn import Urn
from .web import Uri
